@@ -1,20 +1,11 @@
 """Render the concat document to pdf."""
-import datetime as dti
-import difflib
-import hashlib
-import io
 import os
 import pathlib
 import shutil
-import subprocess  # nosec B404
-import sys
 import time
-from typing import Any, Callable, no_type_check
+from typing import no_type_check
 
-import foran.foran as api  # type: ignore
 import yaml
-from foran.report import generate_report  # type: ignore
-from taksonomia.taksonomia import Taxonomy  # type: ignore
 
 import liitos.captions as cap
 import liitos.concat as con
@@ -24,113 +15,34 @@ import liitos.gather as gat
 import liitos.labels as lab
 import liitos.patch as pat
 import liitos.tables as tab
-from liitos import ENCODING, log
+import liitos.tools as too
+from liitos import (
+    ENCODING,
+    FILTER_CS_LIST,
+    FROM_FORMAT_SPEC,
+    LATEX_PAYLOAD_NAME,
+    LOG_SEPARATOR,
+    TOOL_VERSION_COMMAND_MAP,
+    log,
+    parse_csl,
+)
 
 DOC_BASE = pathlib.Path('..', '..')
 STRUCTURE_PATH = DOC_BASE / 'structure.yml'
 IMAGES_FOLDER = 'images/'
 DIAGRAMS_FOLDER = 'diagrams/'
 PATCH_SPEC_NAME = 'patch.yml'
-CHUNK_SIZE = 2 << 15
-TS_FORMAT = '%Y-%m-%d %H:%M:%S.%f +00:00'
-LOG_SEPARATOR = '- ' * 80
 INTER_PROCESS_SYNC_SECS = 0.1
 INTER_PROCESS_SYNC_ATTEMPTS = 10
 
 
-def hash_file(path: pathlib.Path, hasher: Callable[..., Any] | None = None) -> str:
-    """Return the SHA512 hex digest of the data from file."""
-    if hasher is None:
-        hasher = hashlib.sha512
-    hash = hasher()
-    with open(path, 'rb') as handle:
-        while chunk := handle.read(CHUNK_SIZE):
-            hash.update(chunk)
-    return hash.hexdigest()
-
-
-@no_type_check
-def log_subprocess_output(pipe, prefix: str):
-    for line in iter(pipe.readline, b''):  # b'\n'-separated lines
-        cand = line.decode(encoding=ENCODING).rstrip()
-        if cand.strip().strip('[])yex'):
-            if any(
-                [
-                    'microtype' in cand,
-                    'xassoccnt' in cand,
-                    'texlive/2022/texmf-dist/tex/' in cand,
-                    cand == 'erns.sty)',
-                    cand == '(see the transcript file for additional information)',
-                    cand.startswith(r'Overfull \hbox ')
-                    and cand.endswith(r'pt too wide) has occurred while \output is active'),
-                ]
-            ):
-                log.debug(f'{prefix}: %s', cand)
-            else:
-                log.info(f'{prefix}: %s', cand)
-
-
-@no_type_check
-def vcs_probe():
-    """Are we in front, on par, or behind with the upstream?"""
-    try:
-        repo = api.Repo('.', search_parent_directories=True)
-        status = api.Status(repo)
-        api.local_commits(repo, status)
-        api.local_staged(repo, status)
-        api.local_files(repo, status)
-        try:
-            repo_root_folder = repo.git.rev_parse(show_toplevel=True)
-            yield f'Root     ({repo_root_folder})'
-        except Exception:
-            yield 'WARNING - ignored exception when assessing repo root folder location'
-        for line in generate_report(status):
-            yield line.rstrip()
-    except Exception:
-        yield 'WARNING - we seem to not be within a git repository clone'
-
-
-def report_taxonomy(target_path: pathlib.Path) -> None:
-    """Convenience function to report date, size, and checksums of the deliverable."""
-    taxonomy = Taxonomy(target_path, excludes='', key_function='md5')
-    for path in sorted(target_path.parent.rglob('*')):
-        taxonomy.add_branch(path) if path.is_dir() else taxonomy.add_leaf(path)
-    log.info('- Writing render/pdf folder taxonomy to inventory.json ...')
-    taxonomy.dump(sink='inventory', format_type='json', base64_encode=False)
-
-    stat = target_path.stat()
-    size_bytes = stat.st_size
-    mod_time = dti.datetime.fromtimestamp(stat.st_ctime, tz=dti.timezone.utc).strftime(TS_FORMAT)
-    sha612_hash = hash_file(target_path, hashlib.sha512)
-    sha256_hash = hash_file(target_path, hashlib.sha256)
-    sha1_hash = hash_file(target_path, hashlib.sha1)
-    md5_hash = hash_file(target_path, hashlib.md5)
-    log.info('- Ephemeral:')
-    log.info(f'  + name: {target_path.name}')
-    log.info(f'  + size: {size_bytes} bytes')
-    log.info(f'  + date: {mod_time}')
-    log.info('- Characteristic:')
-    log.info('  + Checksums:')
-    log.info(f'    sha512:{sha612_hash}')
-    log.info(f'    sha256:{sha256_hash}')
-    log.info(f'      sha1:{sha1_hash}')
-    log.info(f'       md5:{md5_hash}')
-    log.info('  + Fonts:')
-
-
-@no_type_check
-def unified_diff(left: list[str], right: list[str], left_label: str = 'before', right_label: str = 'after'):
-    """Derive the unified diff between left and right lists of strings as generator of strings."""
-    for line in difflib.unified_diff(left, right, fromfile=left_label, tofile=right_label):
-        yield line.rstrip()
-
-
 @no_type_check
 def der(
     doc_root: str | pathlib.Path, structure_name: str, target_key: str, facet_key: str, options: dict[str, bool | str]
 ) -> int:
     """Later alligator."""
     log.info(LOG_SEPARATOR)
+    log.info('entered render function ...')
     target_code = target_key
     facet_code = facet_key
     if not facet_code.strip() or not target_code.strip():
@@ -139,6 +51,10 @@ def der(
 
     log.info(f'parsed target ({target_code}) and facet ({facet_code}) from request')
 
+    from_format_spec = options.get('from_format_spec', FROM_FORMAT_SPEC)
+    filter_cs_list = parse_csl(options.get('filter_cs_list', FILTER_CS_LIST))
+    log.info(f'parsed from-format-spec ({from_format_spec}) and filters ({", ".join(filter_cs_list)}) from request')
+
     structure, asset_map = gat.prelude(
         doc_root=doc_root, structure_name=structure_name, target_key=target_key, facet_key=facet_key, command='render'
     )
@@ -187,11 +103,7 @@ def der(
 
     log.info(LOG_SEPARATOR)
     log.info('Assessing the local version control status (compared to upstream) ...')
-    log.info(LOG_SEPARATOR)
-    for lines in vcs_probe():
-        for line in lines.split('\n'):
-            log.info(line)
-    log.info(LOG_SEPARATOR)
+    too.ensure_separate_log_lines(too.vcs_probe)
 
     if not STRUCTURE_PATH.is_file() or not STRUCTURE_PATH.stat().st_size:
         log.error(f'render failed to find non-empty structure file at {STRUCTURE_PATH}')
@@ -242,7 +154,7 @@ def der(
             log.info('we will render ...')
         else:
             log.warning('we will not render ...')
-            return 0
+            return 0xFADECAFE
 
         log.info(LOG_SEPARATOR)
         log.info('transforming SVG assets to high resolution PNG bitmaps ...')
@@ -257,18 +169,7 @@ def der(
                 if svg.is_file() and svg.suffix == '.svg':
                     png = str(svg).replace('.svg', '.png')
                     svg_to_png_command = ['svgexport', svg, png, '100%']
-                    process = subprocess.Popen(  # nosec B603
-                        svg_to_png_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
-                    )
-                    with process.stdout:
-                        log_subprocess_output(process.stdout, 'svg-to-png')
-                    return_code = process.wait()
-                    if return_code < 0:
-                        log.error(f'svg-to-png process ({svg_to_png_command}) was terminated by signal {-return_code}')
-                    elif return_code == 0:
-                        log.info(f'svg-to-png process ({svg_to_png_command}) returned {return_code}')
-                    else:
-                        log.error(f'svg-to-png process ({svg_to_png_command}) returned {return_code}')
+                    too.delegate(svg_to_png_command, 'svg-to-png')
 
         special_patching = []
         log.info(LOG_SEPARATOR)
@@ -339,9 +240,10 @@ def der(
             log.info('post-action queue (from reference renaming) is empty - nothing to move')
         log.info(LOG_SEPARATOR)
 
-        fmt_spec = 'markdown+header_attributes+link_attributes+strikeout'
+        fmt_spec = from_format_spec
         in_doc = 'document.md'
-        out_doc = 'document.tex'
+        out_doc = LATEX_PAYLOAD_NAME
+        filters = [added_prefix for expr in filter_cs_list for added_prefix in ('--filter', expr)]
         markdown_to_latex_command = [
             'pandoc',
             '--verbose',
@@ -352,107 +254,64 @@ def der(
             in_doc,
             '-o',
             out_doc,
-            '--filter',
-            'mermaid-filter',
-        ]
+        ] + filters
         log.info(LOG_SEPARATOR)
         log.info(f'executing ({" ".join(markdown_to_latex_command)}) ...')
-        process = subprocess.Popen(  # nosec B603
-            markdown_to_latex_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
-        )
-        with process.stdout:
-            log_subprocess_output(process.stdout, 'markdown-to-latex')
-        return_code = process.wait()
-        if return_code < 0:
-            log.error(
-                f'markdown-to-latex process ({markdown_to_latex_command}) was terminated by signal {-return_code}'
-            )
-        else:
-            log.info(f'markdown-to-latex process ({markdown_to_latex_command}) returned {return_code}')
+        if code := too.delegate(markdown_to_latex_command, 'markdown-to-latex'):
+            return code
 
         log.info(LOG_SEPARATOR)
-        log.info('move any captions below tables ...')
-        with open('document.tex', 'rt', encoding=ENCODING) as handle:
+        log.info(f'load text lines from intermediate {LATEX_PAYLOAD_NAME} file before internal transforms ...')
+        with open(LATEX_PAYLOAD_NAME, 'rt', encoding=ENCODING) as handle:
             lines = [line.rstrip() for line in handle.readlines()]
-        doc_before_caps_patch = 'document-before-caps-patch.tex.txt'
-        with open(doc_before_caps_patch, 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines))
-        lines_caps_patch = cap.weave(lines)
-        with open('document.tex', 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_caps_patch))
-        log.info('diff of the (captions-below-tables) filter result:')
-        log.info(LOG_SEPARATOR)
-        for line in unified_diff(lines, lines_caps_patch):
-            for fine in line.split('\n'):
-                log.info(fine)
-        log.info(LOG_SEPARATOR)
 
-        log.info(LOG_SEPARATOR)
-        log.info('inject stem (derived from file name) labels ...')
-        doc_before_label_patch = 'document-before-inject-stem-label-patch.tex.txt'
-        with open(doc_before_label_patch, 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_caps_patch))
-        lines_inject_stem_label = lab.inject(lines_caps_patch)
-        with open('document.tex', 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_inject_stem_label))
-        log.info('diff of the (inject-stem-derived-labels) filter result:')
-        log.info(LOG_SEPARATOR)
-        for line in unified_diff(lines_caps_patch, lines_inject_stem_label):
-            for fine in line.split('\n'):
-                log.info(fine)
-        log.info(LOG_SEPARATOR)
+        lines = too.execute_filter(
+            cap.weave,
+            head='move any captions below tables ...',
+            backup='document-before-caps-patch.tex.txt',
+            label='captions-below-tables',
+            text_lines=lines,
+        )
 
-        log.info(LOG_SEPARATOR)
-        log.info('scale figures ...')
-        doc_before_figures_patch = 'document-before-scale-figures-patch.tex.txt'
-        with open(doc_before_figures_patch, 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_inject_stem_label))
-        lines_scale_figures = fig.scale(lines_inject_stem_label)
-        with open('document.tex', 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_scale_figures))
-        log.info('diff of the (inject-scale-figures) filter result:')
-        log.info(LOG_SEPARATOR)
-        for line in unified_diff(lines_inject_stem_label, lines_scale_figures):
-            for fine in line.split('\n'):
-                log.info(fine)
-        log.info(LOG_SEPARATOR)
+        lines = too.execute_filter(
+            lab.inject,
+            head='inject stem (derived from file name) labels ...',
+            backup='document-before-inject-stem-label-patch.tex.txt',
+            label='inject-stem-derived-labels',
+            text_lines=lines,
+        )
 
-        log.info(LOG_SEPARATOR)
-        log.info('add options to descriptions (definition lists) ...')
-        doc_before_descriptions_patch = 'document-before-description-options-patch.tex.txt'
-        with open(doc_before_descriptions_patch, 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_scale_figures))
-        lines_description_options = dsc.options(lines_scale_figures)
-        with open('document.tex', 'wt', encoding=ENCODING) as handle:
-            handle.write('\n'.join(lines_description_options))
-        log.info('diff of the (inject-description-options) filter result:')
-        log.info(LOG_SEPARATOR)
-        for line in unified_diff(lines_scale_figures, lines_description_options):
-            for fine in line.split('\n'):
-                log.info(fine)
-        log.info(LOG_SEPARATOR)
+        lines = too.execute_filter(
+            fig.scale,
+            head='scale figures ...',
+            backup='document-before-scale-figures-patch.tex.txt',
+            label='inject-scale-figures',
+            text_lines=lines,
+        )
+
+        lines = too.execute_filter(
+            dsc.options,
+            head='add options to descriptions (definition lists) ...',
+            backup='document-before-description-options-patch.tex.txt',
+            label='inject-description-options',
+            text_lines=lines,
+        )
 
         if options.get('patch_tables', False):
-            log.info(LOG_SEPARATOR)
-            log.info('patching tables EXPERIMENTAL (table-shape) ...')
-            doc_before_table_shape_patch = 'document-before-table-shape-patch.tex.txt'
-            with open(doc_before_table_shape_patch, 'wt', encoding=ENCODING) as handle:
-                handle.write('\n'.join(lines_description_options))
-            lines_table_shape_options = tab.patch(lines_description_options)
-            with open('document.tex', 'wt', encoding=ENCODING) as handle:
-                handle.write('\n'.join(lines_table_shape_options))
-            log.info('diff of the (changed-table-shape) filter result:')
-            log.info(LOG_SEPARATOR)
-            for line in unified_diff(lines_description_options, lines_table_shape_options):
-                for fine in line.split('\n'):
-                    log.info(fine)
-            log.info(LOG_SEPARATOR)
+            lines = too.execute_filter(
+                dsc.options,
+                head='patching tables EXPERIMENTAL (table-shape) ...',
+                backup='document-before-table-shape-patch.tex.txt',
+                label='changed-table-shape',
+                text_lines=lines,
+            )
         else:
             log.info(LOG_SEPARATOR)
             log.info('not patching tables but commenting out (ignoring) any columns command (table-shape) ...')
-            lines_table_shape_options = [
-                f'%IGNORED_{v}' if v.startswith(r'\columns=') else v for v in lines_description_options
-            ]
+            patched_lines = [f'%IGNORED_{v}' if v.startswith(r'\columns=') else v for v in lines]
+            log.info('diff of the (ignore-table-shape-if-not-patched) filter result:')
+            too.log_unified_diff(lines, patched_lines)
+            lines = patched_lines
             log.info(LOG_SEPARATOR)
 
         if need_patching:
@@ -460,20 +319,20 @@ def der(
             log.info('apply user patches ...')
             doc_before_user_patch = 'document-before-user-patch.tex.txt'
             with open(doc_before_user_patch, 'wt', encoding=ENCODING) as handle:
-                handle.write('\n'.join(lines_table_shape_options))
-            lines_user_patches = pat.apply(patches, lines_table_shape_options)
-            with open('document.tex', 'wt', encoding=ENCODING) as handle:
-                handle.write('\n'.join(lines_user_patches))
+                handle.write('\n'.join(lines))
+            patched_lines = pat.apply(patches, lines)
+            with open(LATEX_PAYLOAD_NAME, 'wt', encoding=ENCODING) as handle:
+                handle.write('\n'.join(patched_lines))
             log.info('diff of the (user-patches) filter result:')
-            log.info(LOG_SEPARATOR)
-            for line in unified_diff(lines_table_shape_options, lines_user_patches):
-                for fine in line.split('\n'):
-                    log.info(fine)
-            log.info(LOG_SEPARATOR)
+            too.log_unified_diff(lines, patched_lines)
+            lines = patched_lines
         else:
             log.info(LOG_SEPARATOR)
             log.info('skipping application of user patches ...')
 
+        log.info(LOG_SEPARATOR)
+        log.info(f'Internal text line buffer counts {len(lines)} lines')
+
         log.info(LOG_SEPARATOR)
         log.info('cp -a driver.tex this.tex ...')
         source_asset = 'driver.tex'
@@ -483,45 +342,24 @@ def der(
         latex_to_pdf_command = ['lualatex', '--shell-escape', 'this.tex']
         log.info(LOG_SEPARATOR)
         log.info('1/3) lualatex --shell-escape this.tex ...')
-        process = subprocess.Popen(latex_to_pdf_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # nosec B603
-        with process.stdout:
-            log_subprocess_output(process.stdout, 'latex-to-pdf(1/3)')
-        return_code = process.wait()
-        if return_code < 0:
-            log.error(f'latex-to-pdf process 1/3 ({latex_to_pdf_command}) was terminated by signal {-return_code}')
-        else:
-            log.info(f'latex-to-pdf process 1/3  ({latex_to_pdf_command}) returned {return_code}')
+        if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(1/3)'):
+            return code
 
         log.info(LOG_SEPARATOR)
         log.info('2/3) lualatex --shell-escape this.tex ...')
-        process = subprocess.Popen(latex_to_pdf_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # nosec B603
-        with process.stdout:
-            log_subprocess_output(process.stdout, 'latex-to-pdf(2/3)')
-        return_code = process.wait()
-        if return_code < 0:
-            log.error(f'latex-to-pdf process 2/3 ({latex_to_pdf_command}) was terminated by signal {-return_code}')
-        else:
-            log.info(f'latex-to-pdf process 2/3  ({latex_to_pdf_command}) returned {return_code}')
+        if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(2/3)'):
+            return code
 
         log.info(LOG_SEPARATOR)
         log.info('3/3) lualatex --shell-escape this.tex ...')
-        process = subprocess.Popen(latex_to_pdf_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # nosec B603
-        with process.stdout:
-            log_subprocess_output(process.stdout, 'latex-to-pdf(3/3)')
-        return_code = process.wait()
-        if return_code < 0:
-            log.error(f'latex-to-pdf process 3/3 ({latex_to_pdf_command}) was terminated by signal {-return_code}')
-        else:
-            log.info(f'latex-to-pdf process 3/3  ({latex_to_pdf_command}) returned {return_code}')
+        if code := too.delegate(latex_to_pdf_command, 'latex-to-pdf(3/3)'):
+            return code
 
         if str(options.get('label', '')).strip():
             labeling_call = str(options['label']).strip().split()
             log.info(LOG_SEPARATOR)
             log.info(f'Labeling the resulting pdf file per ({options["label"]})')
-            proc = subprocess.Popen(labeling_call, stdout=subprocess.PIPE)  # nosec B603
-            for line in io.TextIOWrapper(proc.stdout, encoding='utf-8'):
-                print(line.rstrip())
-            sys.stdout.flush()
+            too.delegate(labeling_call, 'label-pdf')
             log.info(LOG_SEPARATOR)
 
         log.info(LOG_SEPARATOR)
@@ -532,17 +370,10 @@ def der(
 
         log.info(LOG_SEPARATOR)
         log.info('Deliverable taxonomy: ...')
-        report_taxonomy(pathlib.Path(target_asset))
+        too.report_taxonomy(pathlib.Path(target_asset))
 
         pdffonts_command = ['pdffonts', target_asset]
-        process = subprocess.Popen(pdffonts_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # nosec B603
-        with process.stdout:
-            log_subprocess_output(process.stdout, '    pdffonts')
-        return_code = process.wait()
-        if return_code < 0:
-            log.error(f'pdffonts process ({pdffonts_command}) was terminated by signal {-return_code}')
-        else:
-            log.info(f'pdffonts process ({pdffonts_command}) returned {return_code}')
+        too.delegate(pdffonts_command, 'assess-pdf-fonts')
 
         log.info(LOG_SEPARATOR)
         log.info('done.')
